qs = require 'qs'
#Consider renaming it to pengines again (like the original)
pengines = require 'pengines'
simpleAuth = require './simpleAuth'
async = require 'async'

class PrologDb
    
    #Map between the ids of the pengines and the pangines themselves. Needs to be kept track of because there needs to be a cleanup to destroy all the slaves
    openQueries = {}
    
    
    constructor : (options) ->
        {@authOptions, @applicationName, @serverUrl, @maxQueryLength, @sourceText} = options
        @maxQueryLength ?= 1000
        @applicationName ?= "swish"
        @sourceText ?= ""
        
        
    open : (callback) =>
        
        if @authOptions?
            simpleAuth @authOptions, (error,...,cookieJar) =>
                @cookieJar = cookieJar
                callback(error,@)
        else
            @cookieJar = request.jar()
            callback(null,@)
    
    
    addPresetSourceText : (sourceText,callback) ->
        @sourceText = if @sourceText? then "#{@sourceText}\n#{fileContent}" else sourceText

    #Perform a connection straight away without going through constructor (convenience method inspired by mongoDB npm module)
    @connect : (options,callback) ->
        #newPrologDb =
        
        new PrologDb(options).open( (error,newPrologDb) =>
            callback(error,newPrologDb)
        )
        
            
            
    query : (queryString,callback) =>
        
        pengineOptions =
            server: @serverUrl,
            #Not sure if it is ideal to have "swish" here, dunno what it really affects
            application: @applicationName,
            #The json-s is necessary to avoid internal server error when retreiving different types, like literal(...)
            format: "json",
            destroy: true,
            sourceText: @sourceText,
            ask: queryString,
            chunk: @maxQueryLength,
            cookieJar: @cookieJar
        
        
        queryResult = null
        
        peng = pengines(pengineOptions)
            .on('error',(obj) -> callback(new Error("Error in operation: #{obj.data}")))
            .on('failure',(error) -> callback(new Error("Pengine object failed: #{obj.data}")))
            .on('destroy', ->
                    callback null,queryResult
                )
            .on('success', (result) ->
                queryResult = result
                peng.destroy()
            )
            
    
    queryMore : (queryId,callback) ->
        openQuery = openQueries[queryId]
        callback ?= openQuery.initialCallback
        
        openQuery.once('success',(result) ->
            callback(null,result,queryId)
        )
        
        openQuery.next()
            
        
    closeQuery : (queryId,callback) ->
        openQueries[queryId].destroy().on('destroy',callback)
    
    #Asks for a number of entries and allow for asking for more entries
    openQuery : (queryString,nEntries,callback) =>
        
        
        peng = pengines(
                server: @serverUrl
                application: @applicationName,
                #The json-s is necessary to avoid internal server error when retreiving different types, like literal(...)
                format: "json",
                destroy: false,
                sourceText: @sourceText,
                ask: queryString,
                chunk: nEntries,
                cookieJar: @cookieJar
            )
            .on('error',(obj) -> callback(new Error("Error in operation: #{obj.data}")))
            .on('failure',(obj) ->
                callback(new Error("Pengine object failed: #{JSON.stringify(obj)}")))
            .on('create', (result) ->
                openQueries[peng.id] = peng
                openQueries[peng.id].initialCallback = callback
            )
            .once('success', (result) ->
                #If there are more results to obtain, attach a callback to get more data
                callback(null,result,peng.id)
            )

        

        
        
    @cleanUp : ->
        for id, peng of openQueries
            if peng.died
                #Todo: investigate if there is a risk for concurrent modification
                delete openQueries[id]
            else
                #TODO: Currently, we don't wait for callback here. Reconsider a synchronous solution?
                peng.destroy()

process.on 'exit', (code) ->
    PrologDb.cleanUp()


module.exports = PrologDb
